#!/usr/bin/env python
# -*- coding: utf-8 -*-

# standard library imports
import gzip
import logging
import os
import sys

# module imports
from .helper import return_filehandle


def count_gff_features(gff):
    counts = {}
    fh = return_filehandle(gff)  # use magic filehandle method
    if not fh:
        logger.error("could not open {}".format(gff))
        sys.exit(1)
    with fh as fopen:
        for line in fopen:
            if not line or line.isspace() or line.startswith("#"):  # skip comments
                continue
            line = line.rstrip()
            f = line.split("\t")  # get fields
            if f[2] not in counts:  # feature type
                counts[f[2]] = 1
                continue
            counts[f[2]] += 1
    fh.close()
    return counts


if __name__ == "__main__":
    print(
        "import me to use stast for genome_main and gene_models_main."
        + "This should be in a class probably... but its w/e for now"
    )
    sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# standard library imports
import errno
import gzip
import logging
import os
import subprocess
import sys


def check_subprocess_dependencies():
    """attempts to run samtools, tabix and bgzip"""
    subprocess.check_call("which bgzip", shell=True)  # check in env
    subprocess.check_call("which tabix", shell=True)  # check in env
    subprocess.check_call("which samtools", shell=True)  # check in env
    subprocess.check_call("which gffread", shell=True)  # check in env
    return True


def check_busco_dependencies():
    """attempts to run samtools, tabix and bgzip"""
    subprocess.check_call("which run_BUSCO.py", shell=True)  # check in env
    subprocess.check_call("which augustus", shell=True)  # check in env
    return True


def setup_logging(log_file, log_level):
    """Return logger based on log_file and log_level"""
    log_level = getattr(logging, log_level.upper(), logging.INFO)
    msg_format = "%(asctime)s|%(name)s|[%(levelname)s]: %(message)s"
    logging.basicConfig(format=msg_format, datefmt="%m-%d %H:%M", level=log_level)
    log_handler = logging.FileHandler(log_file, mode="w")
    formatter = logging.Formatter(msg_format)
    log_handler.setFormatter(formatter)
    logger = logging.getLogger("detect_incongruencies")
    logger.addHandler(log_handler)
    return logger


def return_filehandle(open_me):
    """return file handle for gz compressed or text file"""
    magic_dict = {  # headers for compression
        b"\x1f\x8b\x08": "gz",
        #                 '\x42\x5a\x68': 'bz2',
        #                 '\x50\x4b\x03\x04': 'zip'
    }
    max_bytes = max(len(t) for t in magic_dict)
    with open(open_me, "rb") as f:
        s = f.read(max_bytes)
    for m in magic_dict:
        if s.startswith(m):  # check file header for match with m
            t = magic_dict[m]
            if t == "gz":
                return gzip.open(open_me, "rt")
    #          elif t == 'bz2':
    #              return bz2.open(open_me)
    #          elif t == 'zip':
    #              return zipfile.open(open_me)
    return open(open_me)


def check_file(f):
    """check for file using os.path.isfile"""
    try:
        os.path.isfile(f)
    except OSError:
        raise
    return os.path.isfile(f)


def create_directories(dirpath):
    """make directory path"""
    try:
        os.makedirs(dirpath)
    except OSError as e:
        if e.errno != errno.EEXIST:  # ignore if error is exists else raise
            raise


def change_directories(dirpath):
    """cd to path, does not modify calling process CWD"""
    try:
        os.chdir(dirpath)
    except OSError as e:
        raise


if __name__ == "__main__":
    print("import me to use check_file and return_filehandle." + "This should be in a class probably")
    sys.exit(1)



# -*- coding: utf-8 -*-

# standard library imports
import os
import subprocess
import sys

# third-party imports
import click

# module imports
from . import cli
from .common import logger
from .helper import change_directories
from .helper import check_busco_dependencies
from .helper import create_directories
from .helper import return_filehandle


def preprocess_input(target):
    """Creates a temporary file for BUSCO to act on and sets temp and output"""
    name = os.path.basename(target)
    target_dir = os.path.dirname(target)
    busco_dir = "{}/busco".format(target_dir)
    create_directories(busco_dir)  # make runtime dir
    change_directories(busco_dir)  # go to runtime dir cause busco cannot path
    temp_file = "./{}_buscoready.fa".format(name)
    temp_file_handle = open(temp_file, "w")
    temp_target = return_filehandle(target)
    temp_file_handle.writelines(temp_target)
    temp_file_handle.close()
    return temp_file


def run_busco(busco_target, output, temp_dir, threads, mode, lineage):
    """takes busco_target and performs BUSCO using lineage and mode"""
    cmd = """run_BUSCO.py -f -m {} -c {} -i {} -l {} -t {} -o {}""".format(
        mode, threads, busco_target, lineage, temp_dir, output
    )
    subprocess.check_call(cmd, shell=True)
    return output


@click.command()
@click.option("--target", type=str, help="""TARGETS can be files or directories or both""")
@click.option("--lineage", type=str, help="""BUSCO lineage to compare target against""")
@click.option("--mode", type=str, help="""BUSCO mode (genome, protein, transcript)""")
@click.option("--threads", type=int, default=4, help="""Threads for BUSCO""")
def cli(target, lineage, mode, threads):
    """Determines what type of index to apply to input target"""
    check_busco_dependencies()
    if not (target and lineage):
        logger.error("--target, --lineage and arguments required")
        sys.exit(1)
    target = os.path.abspath(target)
    lineage = os.path.abspath(lineage)
    file_types = ["fna", "fasta", "fa", "faa", "ADD MORE"]
    canonical_types = {
        "genome_main": "genome",
        "mrna": "transcriptome",
        "protein": "proteins",
        "protein_primaryTranscript": "proteins",
    }
    fasta = ["fna", "fasta", "fa", "faa"]
    file_attributes = target.split(".")
    output = False
    if len(file_attributes) < 2:
        error_message = """Target {} does not have a type or attributes.  File must end in either gz, bgz, fasta, fna, fa.""".format(
            target
        )
        logger.error(error_message)
        sys.exit(1)
    if (
        not (file_attributes[-2].lower() in canonical_types or file_attributes[-3].lower() in canonical_types)
        and not mode
    ):
        logger.info(
            "file {} was not recognized as a canonical type and mode cannot be assigned please specify --mode for BUSCO to run anyway".format(
                target
            )
        )
        sys.exit(1)
    if file_attributes[-2].lower() in canonical_types and not mode:
        mode = canonical_types[file_attributes[-2].lower()]
    elif file_attributes[-3].lower() in canonical_types and not mode:
        mode = canonical_types[file_attributes[-3].lower()]
    if not mode:
        logger.info("Mode was not assigned and could not be autodetected, please specify BUSCO mode with --mode")
        sys.exit(1)
    logger.info("Setting up output and making temp input...")
    busco_me = preprocess_input(target)
    name = os.path.basename(target)
    output = "{}_busco".format(name)
    temp_dir = "{}_temp".format(name)
    logger.info("Runnig BUSCO on {} with mode {} and lineage {}...".format(target, mode, lineage))
    output = run_busco(busco_me, output, temp_dir, threads, mode, lineage)
    logger.info("BUSCO done output: {}".format(output))



# -*- coding: utf-8 -*-

# standard library imports
import os
import subprocess
import sys

# third-party imports
import click
from sequencetools.helpers import sequence_helpers

# module imports
from . import cli
from .common import logger
from .helper import check_subprocess_dependencies


def primary_transcript_check(peptides, logger):
    """If the file is not of primary transcripts, selects the longest isoforms.
    """
    seq_handle = open(peptides, "rt")
    longest = {}
    count = 0
    primary = "{}.protein_primaryTranscript.faa".format(".".join(peptides.split(".")[:-2]))
    for record in sequence_helpers.get_seqio_fasta_record(seq_handle):
        count += 1
        my_record = ">{}\n{}".format(record.id, record.seq)
        my_obj = {"record": my_record, "length": len(record.seq)}
        gene_id = ".".join(record.id.split(".")[:-1])
        if gene_id not in longest:
            longest[gene_id] = my_obj
        else:
            if len(record.seq) > longest[gene_id]["length"]:
                longest[gene_id] = my_obj
    if len(longest) == count:
        logger.info("Protein file is already primary transcripts")
        return  # protein file is already primary
    primary_handle = open(primary, "w")
    for r in longest:
        primary_handle.write(longest[r]["record"] + "\n")
    primary_handle.close()
    return primary


def run_gffread(gff, reference):
    """Reads gff3 file and writes mRNA, CDS and Peptides"""
    gff_dir = os.path.dirname(gff)
    gff_attributes = os.path.basename(gff).split(".")
    mrna = "{}/{}.mrna.fna".format(gff_dir, ".".join(gff_attributes[:5]))
    cds = "{}/{}.cds.fna".format(gff_dir, ".".join(gff_attributes[:5]))
    pep = "{}/{}.protein.faa".format(gff_dir, ".".join(gff_attributes[:5]))
    cmd = "gffread {} -g {} -w {} -x {} -y {} -W".format(gff, reference, mrna, cds, pep)
    subprocess.check_call(cmd, shell=True)
    primary_transcript_check(pep)


@cli.command()
@click.option("--target", type=str, help="""GFF file from prefix_gff.""")
@click.option("--reference", type=str, help="""FASTA file from prefix_fasta.""")
def extract_fasta(target, reference):
    """Determines what type of index to apply to input target

    \b
    Example:
        bionorm extract_fasta --target example_jemalong.gff3 --reference
    """
    check_subprocess_dependencies()
    if not (target and reference):
        logger.error("--target and --reference arguments are required")
        sys.exit(1)
    target = os.path.abspath(target)  # get full path
    target_attributes = os.path.basename(target).split(".")
    if os.path.basename(reference).split(".")[-1] == "gz":
        logger.error("GFFREAD cannot process compressed fasta as reference")
        sys.exit(1)
    if len(target_attributes) < 7:
        logger.error("Target file {} is not delimited correctly".format(target))
        sys.exit(1)
    run_gffread(target, reference)  # write readme template



# -*- coding: utf-8 -*-

# standard library imports
import os
import re
import sys
from pathlib import Path

# third-party imports
import click

# module imports
from . import cli
from .common import logger
from .helper import check_file
from .helper import create_directories
from .helper import return_filehandle

# global constants
GENUS_CODE_LEN = 3
SPECIES_CODE_LEN = 2
GFF_SPLIT_MAGIC = ["##gff-version", "3"]


def organism_code(genus, species):
    return f"{genus[:GENUS_CODE_LEN].lower()}{species[:SPECIES_CODE_LEN].lower()}"


def get_species_dirname(genus, species):
    return f"{genus.capitalize()}_{species}"


def get_genome_dir(infra_id, genver=None, annver=None, key=None):
    dirname = f"{infra_id}"
    if genver is not None:
        dirname += f".gnm{genver}"
    if annver is not None:
        dirname += f".ann{annver}"
    if key is not None:
        dirname += f".{key}"
    return dirname


@cli.command()
@click.option("--genver", type=int, required=True, help="Genome version number.")
@click.option("--genus", required=True, help="Genus name of organism.")
@click.option("--species", required=True, help="Species name of organism.")
@click.option("--infra_id", required=True, help="Infraspecies identifier. [e.g., 'A17_HM341']")
@click.option("--key", required=True, metavar="<STRING, len=4>", help="4-character unique identifier.")
@click.argument("fastafile", type=click.Path(exists=True, readable=True, dir_okay=False))
def prefix_fasta(fastafile, genver, genus, species, infra_id, key):
    """Prefix FASTA files to data store standards.

    \b
    Example:
        bionorm prefix-fasta --genver 5 --genus medicago --species truncatula \\
                --infra_id jemalong_A17 --key FAKE example_jemalong.fna
    """
    re_header = re.compile(r"^>(\S+)\s*(.*)")  # match header
    org_code = organism_code(genus, species)
    new_file_dir = Path(".") / get_species_dirname(genus, species) / get_genome_dir(infra_id, genver=genver, key=key)
    new_file_dir.mkdir(parents=True, exist_ok=True)
    fasta_file_path = new_file_dir / f"{org_code}.{infra_id}.gnm{genver}.{key}.genome_main.fna"
    new_fasta = open(fasta_file_path, "w")
    name_prefix = f"{org_code}.{infra_id}.gnm{genver}"
    fh = return_filehandle(fastafile)
    n_headers = 0
    with fh as gopen:
        for line in gopen:
            line = line.rstrip()
            if re_header.match(line):  # header line
                parsed_header = re_header.search(line).groups()
                logger.debug(line)
                logger.debug(parsed_header)
                if isinstance(parsed_header, str):  # check tuple
                    hid = parsed_header
                    new_header = f">{name_prefix}.{hid}"
                    n_headers += 1
                    logger.debug(hid)
                else:
                    if len(parsed_header) == 2:  # header and description
                        hid = parsed_header[0]  # header
                        desc = parsed_header[1]  # description
                        new_header = ">{}.{} {}".format(name_prefix, hid, desc)
                        n_headers += 1
                        logger.debug(hid)
                        logger.debug(desc)
                        logger.debug(new_header)
                    else:
                        logger.error("header {} looks odd...".format(line))
                        sys.exit(1)
                new_fasta.write(new_header + "\n")  # write new header
            else:  # sequence lines
                new_fasta.write(line + "\n")
    new_fasta.close()
    if not n_headers:
        logger.error("file %s contains no headers, are you sure it is a FASTA?", fastafile)
        os.remove(fasta_file_path)
        sys.exit(1)
    if not check_file(fasta_file_path):
        logger.error("Output file {} not found for normalize fasta".format(fasta_file_path))
        sys.exit(1)  # new file not found return False
    logger.debug("%d sequences in output file", n_headers)
    logger.info(fasta_file_path)
    return fasta_file_path


def type_rank(hierarchy, feature_type):
    if feature_type in hierarchy:
        return hierarchy[feature_type]["rank"]
    else:
        return 1000  # no feature id, no children, no order


def update_hierarchy(hierarchy, feature_type, parent_types):
    """Breaks input gff3 line into attributes.

       Determines feature type hierarchy using type and attributes fields
    """
    if not parent_types:  # this feature must be a root
        if feature_type not in hierarchy:  # add to hierarchy
            hierarchy[feature_type] = {"children": [], "parents": [], "rank": 0}
        hierarchy[feature_type]["rank"] = 1  # rank is 1 because no parents
    else:  # feature has parents
        if feature_type not in hierarchy:
            hierarchy[feature_type] = {"children": [], "parents": [], "rank": 0}
        hierarchy[feature_type]["parents"] = parent_types
        for p in parent_types:  # set children and aprents for all p
            if hierarchy.get(p):
                if feature_type not in hierarchy[p]["children"]:
                    hierarchy[p]["children"].append(feature_type)
            else:
                hierarchy[p] = {"children": [feature_type], "parents": [], "rank": 0}


@cli.command()
@click.option("--gnm", required=True, type=int, help="Genome version number.")
@click.option("--ann", required=True, type=int, help="""Annotation version number.""")
@click.option("--genus", required=True, help="Genus name of organism.")
@click.option("--species", required=True, help="Species name of organism.")
@click.option("--infra_id", required=True, help="Infraspecies identifier. [e.g., 'A17_HM341']")
@click.option("--key", required=True, metavar="<STRING, len=4>", help="4-character unique identifier.")
@click.option("--sort_only", is_flag=True, help="Perform sorting only.", default=False)
@click.argument("gff3file", type=click.Path(exists=True, readable=True, dir_okay=False))
def prefix_gff(gff3file, gnm, ann, genus, species, infra_id, key, sort_only):
    """Prefix and sort GFF3 file to data store standards.

    \b
    Example:
        bionorm prefix-gff --gnm 5 --ann 1 --species truncatula --genus medicago \\
                --infra_id jemalong_A17 --key FAKE example_jemalong.fna
    """
    gnm = "gnm{}".format(gnm)
    ann = "ann{}".format(ann)
    if sort_only:
        logger.debug("sorting ONLY...")
    else:
        logger.debug("prefixing and sorting...")
    org_code = organism_code(genus, species)
    new_gff_name = f"{org_code}.{infra_id}.{gnm}.{ann}.{key}.gene_models_main.gff3"
    species_dir = Path(".") / get_species_dirname(genus, species)
    new_file_dir = species_dir / "{}.{}.{}.{}".format(infra_id, gnm, ann, key)
    create_directories(new_file_dir)  # make genus species dir for output
    gff_file_path = new_file_dir / new_gff_name
    new_gff = open(gff_file_path, "w")
    get_id = re.compile("ID=([^;]+)")  # gff3 get id string
    get_name = re.compile("Name=([^;]+)")  # get name string
    get_parents = re.compile("Parent=([^;]+)")  # get parents
    gff3_lines = []  # will be sorted after loading and hierarchy
    sub_tree = {}  # feature hierarchy sub tree
    type_hierarchy = {}  # type hierarchy for features, will be ranked
    prefix_name = False
    fh = return_filehandle(gff3file)  # get filehandle for gff3file
    n_lines = 0
    with fh as gopen:
        for line in gopen:
            line = line.rstrip()
            if not line:
                continue
            n_lines += 1
            if n_lines == 1:  # magic in first line
                if not line.split() == GFF_SPLIT_MAGIC:
                    logger.error("File does not start with GFF3 magic, are you sure is is a GFF3?")
                    new_gff.close()
                    os.remove(gff_file_path)
                    sys.exit(1)
            if line.startswith("#"):  # header
                if sort_only:  # do not prefix
                    new_gff.write("{}\n".format(line))
                    continue
                if line.startswith("##sequence-region"):  # replace header
                    fields = re.split(r"\s+", line)
                    ref_name = "{}.{}.{}.{}".format(org_code, infra_id, gnm, fields[1])
                    line = re.sub(fields[1], ref_name, line)
                new_gff.write("{}\n".format(line))
                continue
            fields = line.split("\t")
            gff3_lines.append(fields)
            feature_id = get_id.search(fields[-1])
            parent_ids = get_parents.search(fields[-1])
            if parent_ids:
                parent_ids = parent_ids.group(1).split(",")
            else:
                parent_ids = []
            if not feature_id:
                continue
            feature_id = feature_id.group(1)
            sub_tree[feature_id] = {"type": fields[2], "parent_ids": parent_ids}  # parent child
    for f in sub_tree:
        feature_type = sub_tree[f]["type"]
        parent_types = []
        for p in sub_tree[f]["parent_ids"]:
            parent_type = sub_tree[p]["type"]
            if not parent_type:  # must have type
                logger.error("could not find type for parent {}".format(parent_type))
                sys.exit(1)  # error no type
            parent_types.append(parent_type)
        update_hierarchy(type_hierarchy, feature_type, parent_types)
    ranking = 1  # switch to stop ranking in while below
    while ranking:  # this is over-engineered for tabix indexing, but tis cool
        check = 0  # switch to indicate no features unranked
        for t in sorted(type_hierarchy, key=lambda k: type_hierarchy[k]["rank"], reverse=True):
            if type_hierarchy[t]["rank"]:  # rank !=0
                for c in type_hierarchy[t]["children"]:
                    if not type_hierarchy[c]["rank"]:
                        type_hierarchy[c]["rank"] = type_hierarchy[t]["rank"] + 1
            else:  # rank == 0
                check = 1
        if not check:  # escape loop when all features are ranked
            ranking = 0
    feature_prefix = "{}.{}.{}.{}".format(org_code, infra_id, gnm, ann)
    for l in sorted(
        gff3_lines, key=lambda x: (x[0], int(x[3]), type_rank(type_hierarchy, x[2]))
    ):  # rank by chromosome, start, type_rank and stop
        if sort_only:  # do not prefix or change IDs
            l = "\t".join(l)
            new_gff.write("{}\n".format(l))
            continue
        l[0] = "{}.{}.{}.{}".format(org_code, infra_id, gnm, l[0])  # rename
        l = "\t".join(l)
        feature_id = get_id.search(l)
        feature_name = get_name.search(l)
        feature_parents = get_parents.search(l)
        if feature_id:  # if id set new id
            new_id = "{}.{}".format(feature_prefix, feature_id.group(1))
            l = get_id.sub("ID={}".format(new_id), l)
        if feature_name and prefix_name:  # if name and flag set new name
            new_name = "{}.{}".format(feature_prefix, feature_name.group(1))
            l = get_name.sub("Name={}".format(new_name), l)
        if feature_parents:  # parents set new parent ids
            parent_ids = feature_parents.group(1).split(",")
            new_ids = []
            for p in parent_ids:  # for all parents
                new_id = "{}.{}".format(feature_prefix, p)
                new_ids.append(new_id)
            new_ids = ",".join(new_ids)  # set delimiter for multiple parents
            l = get_parents.sub("Parent={}".format(new_ids), l)
        new_gff.write("{}\n".format(l))
    new_gff.close()
    if not check_file(gff_file_path):  # check for output error if not found
        logger.error("Output file {} not found for normalize gff".format(gff_file_path))
        sys.exit(1)  # new file not found return False
    logger.info(gff_file_path)
    return gff_file_path  # return path to new gff file



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# standard library imports
import gzip
import logging
import os
import re
import subprocess
import sys

# third-party imports
import click
from ruamel.yaml import YAML
from ruamel.yaml import RoundTripDumper

# module imports
from ..helper import check_file
from ..helper import check_subprocess_dependencies
from ..helper import create_directories
from ..helper import return_filehandle
from ..helper import setup_logging


def read_yaml(template):
    """Reads yaml file and checks for valid yaml"""
    yaml = YAML()
    target_yaml = yaml.load(open(template, "rt"))
    return yaml, target_yaml


def generate_readme(template, attributes, logger):
    """Writes target YAML file"""
    yaml, my_yaml = read_yaml(template)
    my_readme = "{}/README.{}.yaml".format(attributes["target_dir"], attributes["key"])
    identifier = attributes["key"]
    sci_name = "{} {}".format(attributes["genus"].capitalize(), attributes["species"].lower())
    sci_name_abr = attributes["gensp"].lower()
    genotype = attributes["infra_id"]
    my_yaml["identifier"] = identifier  # set key
    my_yaml["genotype"] = genotype  # set infraspecific id
    my_yaml["scientific_name"] = sci_name  # Genus species
    my_yaml["scientific_name_abbrev"] = sci_name_abr  # gensp
    readme_handle = open(my_readme, "w")
    yaml.dump(my_yaml, readme_handle)
    #    for m in my_yaml:
    readme_handle.close()


@click.command()
@click.option("--target", type=str, help="""Formatted Directory from Previous Normalizer Steaps""")
@click.option("--template", type=str, help="""YAML file to be used as the template for the README""")
@click.option(
    "--log_file",
    metavar="<FILE>",
    default="./normalizer_index.log",
    help="""File to write log to. (default:./normalizer_index.log)""",
)
@click.option(
    "--log_level",
    metavar="<LOGLEVEL>",
    default="INFO",
    help="""Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)""",
)
def cli(target, template, log_file, log_level):
    """Determines what typae of index to apply to input target"""
    logger = setup_logging(log_file, log_level)
    if not (target and template):
        logger.error("--target and --template arguments are required")
        sys.exit(1)
    target = os.path.abspath(target)  # get full path
    if not os.path.isdir(target):
        logger.error("target dmust be a directory")
        sys.exit(1)
    organism_dir = os.path.basename(os.path.dirname(target))
    target_dir = os.path.basename(target)
    organism_attributes = organism_dir.split("_")  # Genus_species
    target_attributes = target_dir.split(".")
    if len(organism_attributes) != 2:
        logger.error("Parent directory {} is not Genus_species".format(organism_dir))
        sys.exit(1)
    if len(target_attributes) < 3:
        logger.error("Target directory {} is not delimited correctly".format(target_dir))
        sys.exit(1)
    # abbreviation in yaml README file
    gensp = "{}{}".format(organism_attributes[0][:3], organism_attributes[1][:2])
    attributes = {
        "genus": organism_attributes[0],
        "species": organism_attributes[1],
        "gensp": gensp,
        "key": target_attributes[-1],
        "infra_id": target_attributes[0],
        "target_dir": target,
    }
    #    print('{}\t{}\t{}'.format(template, attributes, logger))
    generate_readme(template, attributes, logger)  # write readme template



# -*- coding: utf-8 -*-
# standard library imports
import json
import logging
import os
import re
import subprocess
import sys
from glob import glob

# third-party imports
import click
from sequencetools.tools.basic_fasta_stats import basic_fasta_stats

# module imports
from . import specification_checks
from .helper import check_file
from .metrics import count_gff_features


class Detector:
    """Class to detect datastore file incongruencies with

       https://github.com/LegumeFederation/datastore
    """

    def __init__(self, target, **kwargs):
        """Check for logger, check for gt"""
        subprocess.check_call("which gt", shell=True)  # check for gt in env
        self.options = kwargs
        self.checks = {}  # object that determines which checks are skipped
        self.checks["genome_main"] = kwargs.get("disable_genome_main", True)
        self.checks["gene_models_main"] = kwargs.get("disable_gene_models_main", True)
        self.checks["perform_gt"] = kwargs.get("disable_gt", True)
        self.checks["fasta_headers"] = kwargs.get("disable_fasta_headers", True)
        self.no_nodes = kwargs.get("no_nodes", False)
        self.canonical_types = [
            "genome_main",
            "protein_primaryTranscript",
            "protein",
            "gene_models_main",
            "ADDMORESTUFF",
        ]  # types for detector
        self.canonical_parents = {
            "genome_main": None,
            "gene_models_main": "genome_main",
            "protein_primaryTranscript": "gene_models_main",
            "protein": "gene_models_main",
        }
        self.rank = {"genome_main": 0, "gene_models_main": 1, "protein": 2, "protein_primaryTranscript": 2}
        self.log_level = kwargs.get("log_level", "INFO")
        self.log_file = kwargs.get("log_file", "./incongruencies.log")
        self.output_prefix = kwargs.get("output_prefix", "./incongruencies")
        self.write_me = {}
        self.passed = {}  # dictionary of passing names
        self.target_objects = {}  # store all target pairings self.get_targets
        self.fasta_ids = {}
        self.reporting = {}
        self.node_data = {}  # nodes for DSCensor
        self.target = os.path.abspath(target)
        self.target_readme = ""
        self.domain = "https://legumeinfo.org/data/public"
        self.setup_logging()  # setup logging sets self.logger
        logger = self.logger
        self.get_target_type()  # sets self.target_name and self.target_type
        if not self.target_type:  # target type returned False not recognized
            logger.error("Target type not recognized for {}".format(self.target))
            sys.exit(1)
        # if not os.environ.get('BUSCO_ENV_FILE', None) and not kwargs.get('no_busco'):
        #    logger.error('''
        #        BUSCO_ENV_File Must Be set, please export
        #
        #                This is used to source an Environment for BUSCO as the BUSCO
        #                Config Does Still Requires AUGUSTUS Environment variables be set
        #                ''')
        logger.info("Target type looks like {}".format(self.target_type))
        self.get_targets()
        #        logger.debug(''.format(self.target_objects))
        logger.info("Performing Checks for the Following:\n")
        for t in self.target_objects:  # for each object set validate
            logger.debug("{}".format(t))
            logger.debug("{}".format(self.target_objects[t]))
            logger.info("Parent {}:".format(t))
            count = 0
            set_primary = ""
            primary = False
            for c in self.target_objects[t]["children"]:
                logger.info("Child {}".format(c))
                if self.target_objects[t]["children"][c]["type"] == "protein_primaryTranscript":
                    primary = True
                if self.target_objects[t]["children"][c]["type"] == "protein":
                    count += 1
                    set_primary = c
            if count == 1 and not primary:
                self.target_objects[t]["children"][set_primary]["type"] = "protein_primaryTranscript"
                self.target_objects[t]["children"][set_primary]["node_data"][
                    "canonical_type"
                ] = "protein_primaryTranscript"
            if count > 1 and not primary:
                logger.error("Multiple protein files found for {} one must be renamed to primary".format(t))
                sys.exit(1)
        #            for c in self.target_objects[t]['children']:
        #                logger.info('{}'.format(c))
        logger.info("Initialized Detector\n")

    def setup_logging(self):
        """Return logger based on log_file and log_level"""
        log_level = getattr(logging, self.log_level.upper(), logging.INFO)
        msg_format = "%(asctime)s|%(name)s|[%(levelname)s]: %(message)s"
        logging.basicConfig(format=msg_format, datefmt="%m-%d %H:%M", level=log_level)
        log_handler = logging.FileHandler(self.log_file, mode="w")
        formatter = logging.Formatter(msg_format)
        log_handler.setFormatter(formatter)
        logger = logging.getLogger("detect_incongruencies")
        logger.addHandler(log_handler)
        self.logger = logger

    def get_target_type(self):
        """Determines what type of target this is, is it a file or directory?

           if its a directory, is it an organism directory or a data directory?
        """
        logger = self.logger
        target_name = os.path.basename(self.target)
        self.target_name = target_name
        self.target_type = False
        logger.debug(self.target_name)
        if target_name.endswith(".gz"):  # all datastore files end in .gz
            self.target_type = "file"
        elif len(target_name.split("_")) == 2 and len(target_name.split(".")) < 3:
            self.target_type = "organism_dir"  # will always be Genus_species
        elif len(target_name.split(".")) >= 3:  # standard naming minimum
            self.target_type = "data_dir"

    def get_targets(self):
        """Gets and discovers target files relation to other files

           if the target is a directory, the program will discover

           all related files that can be checked
        """
        logger = self.logger
        target = self.target
        target_type = self.target_type
        if target_type == "file":  # starting with a file
            self.add_target_object()
            return
        elif target_type == "data_dir" or target_type == "organism_dir":
            self.get_all_files()  # works for both data and organism
            return
        # elif target_type == 'organism_dir':  # entire organism
        #    self.get_all_files()
        #    return

    def get_all_files(self):
        """Walk down filetree and recursively return all files

           These will be checked using add_target_object
        """
        logger = self.logger
        target = self.target
        for root, directories, filenames in os.walk(target):
            for filename in filenames:  # we only care about the files
                my_target = os.path.join(root, filename)  # make path
                logger.debug("Checking file {}".format(my_target))
                self.target = my_target  # set target
                self.target_name = os.path.basename(self.target)
                self.add_target_object()  # add target if canonical

    def get_target_file_type(self, target_file):
        """Determines if file is fasta, gff3, vcf, etc"""
        file_type = target_file.split(".")[-2].lower()
        if file_type == "fna" or file_type == "faa" or file_type == "fasta":
            file_type = "fasta"
        elif file_type == "gff" or file_type == "gff3":
            file_type = "gff3"
        else:  # add more
            return False
        return file_type

    def add_target_object(self):
        """Uses parent child logic to create a datastructure for objects"""
        logger = self.logger
        target = self.target
        logger.debug(target)
        if not check_file(target):
            logger.error("Could not find file: {}".format(target))
            sys.exit(1)
        target_attributes = self.target_name.split(".")
        if len(target_attributes) < 3 or self.target_name[0] == "_":
            logger.debug("File {} does not seem to have attributes".format(target))
            return
        canonical_type = target_attributes[-3]  # check content type
        if canonical_type not in self.canonical_types:  # regject
            logger.debug("Type {} not recognized in {}.  Skipping".format(canonical_type, self.canonical_types))
            return
        organism_dir_path = os.path.dirname(os.path.dirname(target))  # org dir
        organism_dir = os.path.basename(os.path.dirname(os.path.dirname(target)))  # org dir
        target_dir = os.path.basename(os.path.dirname(target))
        genus = organism_dir.split("_")[0].lower()
        species = organism_dir.split("_")[1].lower()
        target_format = target_attributes[-2]  # get gff, fna, faa all gx
        target_key = target_attributes[-4]  # get key
        target_ref_type = self.canonical_parents[canonical_type]
        logger.debug("Getting target files reference if necessary...")
        file_type = self.get_target_file_type(self.target_name)
        file_url = "{}/{}/{}/{}".format(self.domain, organism_dir, target_dir, self.target_name)
        target_node_object = {
            "filename": self.target_name,
            "filetype": file_type,
            "canonical_type": canonical_type,
            "url": file_url,
            "counts": "",
            "genus": genus,
            "species": species,
            "origin": "LIS",
            "infraspecies": target_attributes[1],
            "derived_from": [],
            "child_of": [],
        }
        if len(target_attributes) > 7 and target_ref_type:  # check parent
            logger.debug("Target Derived from Some Reference Searching...")
            ref_glob = "{}/{}*/*{}.*.gz".format(organism_dir_path, ".".join(target_attributes[1:3]), target_ref_type)
            if self.rank[canonical_type] > 1:  # feature has a subtype
                ref_glob = "{}/{}*/*{}.*.gz".format(
                    organism_dir_path, ".".join(target_attributes[1:4]), target_ref_type
                )
            my_reference = self.get_reference(ref_glob)
            if my_reference not in self.target_objects:  # new parent
                parent_name = os.path.basename(my_reference)
                file_type = self.get_target_file_type(parent_name)
                organism_dir = os.path.basename(os.path.dirname(os.path.dirname(my_reference)))
                ref_dir = os.path.basename(os.path.dirname(my_reference))
                file_url = "{}/{}/{}/{}".format(self.domain, organism_dir, ref_dir, parent_name)
                ref_node_object = {
                    "filename": parent_name,
                    "filetype": file_type,
                    "canonical_type": target_ref_type,
                    "url": file_url,
                    "counts": "",
                    "genus": genus,
                    "species": species,
                    "origin": "LIS",
                    "infraspecies": target_attributes[1],
                    "derived_from": [],
                    "child_of": [],
                }
                target_node_object["child_of"].append(parent_name)
                target_node_object["derived_from"].append(parent_name)
                self.target_objects[my_reference] = {
                    "type": target_ref_type,
                    "node_data": ref_node_object,
                    "readme": "",
                    "children": {},
                }
                self.target_objects[my_reference]["children"][target] = {
                    "node_data": target_node_object,
                    "type": canonical_type,
                }
            else:  # the parent is already in the data structure add child
                if target not in self.target_objects[my_reference]["children"]:
                    parent_name = os.path.basename(my_reference)
                    target_node_object["child_of"].append(parent_name)
                    target_node_object["derived_from"].append(parent_name)
                    self.target_objects[my_reference]["children"][target] = {
                        "node_data": target_node_object,
                        "type": canonical_type,
                    }
        else:  # target is a reference
            if target_ref_type:
                logger.error("Reference was not found or file has <=7 fields")
                sys.exit(1)
            logger.debug("Target has no Parent, it is a Reference")
            if target not in self.target_objects:
                self.target_objects[target] = {"type": canonical_type, "node_data": target_node_object, "children": {}}

    def get_reference(self, glob_target):
        """Finds the FASTA reference for some prefix"""
        logger = self.logger
        if len(glob(glob_target)) > 1:  # too many references....?
            logger.error("Multiple references found {}".format(glob_target))
            sys.exit(1)
        reference = glob(glob_target)
        if not reference:  # if the objects parent could not be found
            logger.error("Could not find ref glob: {}".format(glob_target))
            sys.exit(1)
        reference = glob(glob_target)[0]
        if not os.path.isfile(reference):  # if cannot find reference file
            logger.error("Could not find main target {}".format(reference))
            sys.exit(1)
        logger.debug("Found reference {}".format(reference))
        return reference

    def write_node_object(self):
        """Write object node for loading into DSCensor

           file name is object name from json
        """
        my_name = self.write_me["filename"]
        if self.write_me["canonical_type"] == "genome_main":
            self.write_me["counts"] = basic_fasta_stats(self.target, 10, False)
        elif self.write_me["canonical_type"] == "gene_models_main":
            self.write_me["counts"] = count_gff_features(self.target)
        my_file = open("./{}.json".format(my_name), "w")
        my_file.write(json.dumps(self.write_me))
        my_file.close()

    def run_busco(self, mode, file_name):
        """Runs BUSCO using BUSCO_ENV_FILE env var and input

           and outputs to file_name
        """
        logger = self.logger  # set logging
        node_data = self.node_data  # get current targets nodes
        busco_parse = re.compile(r"C:(.+)\%\[S:(.+)\%,D:(.+)\%\],F:(.+)\%,M:(.+)\%,n:(\d+)")
        output = "{}.busco".format(".".join(file_name.split(".")[:-2]))
        cmd = "run_BUSCO.py --mode {} --lineage {}".format(mode, "lineage")
        outdir = "./run_{}".format(output)  # output from BUSCO
        short_summary = glob(outdir + "/short_summary*.busco.txt")
        if not short_summary:
            logger.error("BUSCO short summary not found for {}".format(outdir))
            sys.exit(1)
        short_summary = short_summary[0]
        with open(short_summary) as fopen:
            for line in fopen:
                line = line.rstrip()
                if line.startswith("#") or not line:
                    continue
                if line.startswith("\tC:"):
                    line = line.replace("\t", "")
                    percentages = busco_parse.search(line)  # read summary
                    total = int(percentages.group(6))
                    complete = float(percentages.group(1))
                    frag = float(percentages.group(4))
                    single = float(percentages.group(2))
                    duplicate = float(percentages.group(3))
                    missing = float(percentages.group(5))
                    node_data["busco"] = {
                        "total_buscos": total,  # node BUSCO
                        "complete_buscos": complete,
                        "fragmented_buscos": frag,
                        "single_copy_buscos": single,
                        "duplicate_buscos": duplicate,
                        "missing_buscos": missing,
                    }

    def detect_incongruencies(self):
        """Detects all incongruencies in self.target_objects

           report all incongruencies to self.output_prefix

           Probably refactor this at some point to use a self.dict

           This is to allow just iteration and lookup to occur and remove

           basically all this conditional foo
        """
        logger = self.logger
        targets = self.target_objects  # get objects from class init
        no_nodes = self.no_nodes  # if true, no nodes for DSCensor written
        for reference in sorted(targets, key=lambda k: self.rank[targets[k]["type"]]):
            #            logger.info('HERE {}'.format(reference))
            if reference not in self.passed:
                self.passed[reference] = 0
                self.target = reference
                ref_method = getattr(
                    specification_checks, targets[reference]["type"]  # reads checks from spec
                )  # type ex genome_main
                if not ref_method:  # if the target isnt in the hierarchy continue
                    logger.debug("Check for {} does not exist".format(targets[reference]["type"]))
                    continue
                logger.debug(ref_method)
                my_detector = ref_method(self, **self.options)
                passed = my_detector.run()
                if passed:  # validation passed writing object node for DSCensor
                    self.passed[reference] = 1
                    self.node_data = targets[reference]["node_data"]
                    #                if targets[reference]['type'] == 'genome_main':  # BUSCO
                    #                    file_name = targets[reference]['node_data']['filename']
                    #                    if not self.options.get('no_busco'):
                    #                        self.run_busco('genome', file_name)
                    if not no_nodes:
                        logger.info("Writing node object for {}".format(reference))
                        # dscensor node
                        self.write_me = targets[reference]["node_data"]
                        self.write_node_object()  # write node for dscensor loading
                logger.debug("{}".format(targets[reference]))
            if self.target_objects[reference]["children"]:  # process children
                children = self.target_objects[reference]["children"]
                for c in children:
                    if c in self.passed:
                        logger.debug("Child {} Already Passed".format(c))
                        continue

                    self.passed[c] = 0
                    #                    logger.info('HERE child {}'.format(c))
                    logger.info("Performing Checks for {}".format(c))
                    self.target = c
                    child_method = getattr(
                        specification_checks, children[c]["type"]  # check for spec
                    )  # exgene_models_main
                    if not child_method:
                        logger.warning("Check for {} does not exist".format(children[c]["type"]))
                        continue
                    logger.debug(child_method)
                    my_detector = child_method(self, **self.options)
                    passed = my_detector.run()
                    if passed:  # validation passed writing object node for DSCensor
                        self.passed[c] = 1
                        if not no_nodes:
                            logger.info("Writing node object for {}".format(c))
                            self.write_me = children[c]["node_data"]
                            self.write_node_object()
                    logger.debug("{}".format(c))


@click.command()
@click.option("--target", multiple=True, help="""TARGETS can be files or directories or both""")
@click.option("--no_busco", is_flag=True, help="""Disable BUSCO""")
@click.option("--no_nodes", is_flag=True, help="""Disable DSCensor Stats and Node Generation""")
@click.option(
    "--log_file", metavar="<FILE>", default="./bionorm.log", help="""File to write log to. (default:./bionorm.log)"""
)
@click.option(
    "--log_level",
    metavar="<LOGLEVEL>",
    default="INFO",
    help="""Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)""",
)
# @click.option('--normalize', is_flag=True,
#             help='''Normalizes provided files.
# Incongruencies in FASTA will be corrected if the provided genome name
# passes checks.
# The gff file will be tidied if it fails gff3 validation in gt:
#    gt gff3 -sort -tidy -retainids input.gff3 > out.gff3
# ''')
def cli(target, no_busco, no_nodes, log_file, log_level):
    """incongruency_detector:

        Detect Incongruencies with LIS Data Store Standard
    """
    if not target:
        print("Must specify at least one --target.  Run with --help for usage")
        sys.exit(1)
    data_store_home = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    templates_dir = "{}/templates".format(data_store_home)
    readme_template = "{}/template__README.KEY.yml".format(templates_dir)
    options = {
        "log_level": log_level,
        "log_file": log_file,
        "no_busco": no_busco,
        "no_nodes": no_nodes,
        "readme_template": readme_template,
    }
    for t in target:
        detector = Detector(t, **options)  # initializers
        detector.detect_incongruencies()  # class method runs all detection methods


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# standard library imports
import gzip
import logging
import os
import re
import subprocess
import sys

# third-party imports
import click
from ruamel.yaml import YAML

# module imports
from ..helper import check_file
from ..helper import check_subprocess_dependencies
from ..helper import create_directories
from ..helper import return_filehandle
from ..helper import setup_logging


def index_fasta(fasta):
    """Indexes the supplied fasta file using bgzip and samtools"""
    compressed_fasta = "{}.gz".format(fasta)  # will now have gz after bgzip
    if not check_file(compressed_fasta):
        cmd = "bgzip -f --index {}".format(fasta)  # bgzip command
        subprocess.check_call(cmd, shell=True)
    cmd = "samtools faidx {}".format(compressed_fasta)  # samtools faidx
    subprocess.check_call(cmd, shell=True)
    return compressed_fasta  # return new compressed and indexed gff


def index_gff3(gff3):
    """Indexes the supplied gff3 file using bgzip and tabix"""
    compressed_gff = "{}.gz".format(gff3)  # will now have gz after bgzip
    if not check_file(compressed_gff):
        cmd = "bgzip -f {}".format(gff3)  # bgzip command
        subprocess.check_call(cmd, shell=True)
    cmd = "tabix -p gff {}".format(compressed_gff)  # tabix index command
    standard_outval = subprocess.call(cmd.split(" "))
    if standard_outval:
        # tabix index command
        cmd = "tabix --csi -p gff {}".format(compressed_gff)
        subprocess.check_call(cmd, shell=True)
    return compressed_gff  # return new compressed and indexed gff


@click.command()
@click.option("--target", type=str, help="""TARGETS can be files or directories or both""")
@click.option(
    "--log_file",
    metavar="<FILE>",
    default="./normalizer_index.log",
    help="""File to write log to. (default:./normalizer_index.log)""",
)
@click.option(
    "--log_level",
    metavar="<LOGLEVEL>",
    default="INFO",
    help="""Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)""",
)
def cli(target, log_file, log_level):
    """Determines what type of index to apply to input target"""
    check_subprocess_dependencies()
    logger = setup_logging(log_file, log_level)
    if not target:
        logger.error("--target argument required")
        sys.exit(1)
    target = os.path.abspath(target)
    file_types = ["fna", "fasta", "fa", "gff", "gff3", "faa", "ADD MORE"]
    fasta = ["fna", "fasta", "fa", "faa"]
    gff3 = ["gff", "gff3"]
    file_attributes = target.split(".")
    new_file = False
    if len(file_attributes) < 2:
        error_message = """Target {} does not have a type or attributes.  File must end in either gz, bgz, fasta, fna, fa, gff, or gff3.""".format(
            target
        )
        logger.error(error_message)
        sys.exit(1)
    file_type = file_attributes[-1]
    if file_type == "gz" or file_type == "bgz":
        file_type = file_attributes[-2]
        logger.error("Uncompress file for indexing.  Compression is done specifically for each index type")
        sys.exit(1)
    if file_type not in file_types:
        logger.error("File {} is not a type in {}".format(target, file_types))
        sys.exit(1)
    if file_type in fasta:
        logger.info("Target is a FASTA file indexing...")
        new_file = index_fasta(target)
        logger.info("Indexing done, final file: {}".format(new_file))
    if file_type in gff3:
        logger.info("Target is a gff3 file indexing...")
        new_file = index_gff3(target)
        logger.info("Indexing done, final file: {}".format(new_file))
    if not new_file:
        logger.error("Indexing FAILED.  See Log.")
        sys.exit(1)
    return new_file



# -*- coding: utf-8 -*-
"""Global constants and common helper functions.
"""
# standard library imports
import locale
import logging

# third-party imports
import click

#
# global constants
#
PROGRAM_NAME = "bionorm"
CONFIG_FILE_ENVVAR = "BIONORM_CONFIG_FILE_PATH"
#
# global logger object
#
logger = logging.getLogger(PROGRAM_NAME)
#
# set locale so grouping works
#
for localename in ["en_US", "en_US.utf8", "English_United_States"]:
    try:
        locale.setlocale(locale.LC_ALL, localename)
        break
    except BaseException:
        continue

#
# helper functions used in multiple places
#
def get_user_context_obj():
    """Returns the user context, containing logging and configuration data.

    :return: User context object (dict)
    """
    return click.get_current_context().obj



# -*- coding: utf-8 -*-
"""
bionorm -- normalize and verity genomic data files
"""

# This file makes it easier for developers to test in-place via the command
# python3 -m bionorm
# from the directory above this one.

# module imports
from .__init__ import cli

if __name__ == "__main__":
    cli(auto_envvar_prefix="BIONORM")



# -*- coding: utf-8 -*-
"""bionorm -- normalize and validate genomic data files.

This tool is used to ingest data into the Legume Federation Data Store
(https://www.legumefederation.org/en/data-store/) and is useful anytime
genomes must be formatted and checked for consistency.
"""

# standard library imports
import warnings
from pkg_resources import iter_entry_points

# third-party imports
import click
from click_plugins import with_plugins

# module imports
from .cli import Logging_CLI_Builder
from .common import PROGRAM_NAME
from .common import logger

# global constants
DEFAULT_FIRST_N = 0


# define parser options
parser_options = [
    {
        "args": ("--progress",),
        "kwargs": {"is_flag": True, "show_default": True, "default": False, "help": "Show a progress bar."},
    },
    {"args": ("--first_n",), "kwargs": {"default": 0, help: "Process only this many records. [default: all]"}},
    {
        "args": ("--warnings_as_errors", "-e"),
        "kwargs": {"is_flag": True, "show_default": True, "default": False, "help": "Warnings cause exceptions."},
    },
]


def bionorm_check(warnings_as_errors=False, **others):
    """CLI check function, parameters must be keyworded."""
    if warnings_as_errors:
        logger.warn("Runtime warnings (e.g., from pandas) will cause exceptions")
        warnings.filterwarnings("error")


# define the CLI
cli_builder = Logging_CLI_Builder(PROGRAM_NAME, logger, global_options_list=parser_options)

# create CLI
@with_plugins(iter_entry_points(PROGRAM_NAME + ".cli_plugins"))
@click.group(epilog=cli_builder.author + " <" + cli_builder.email + ">.  " + cli_builder.copyright)
@click.option("-v", "--verbose", is_flag=True, show_default=True, default=False, help="Log debugging info to stderr.")
@click.option("-q", "--quiet", is_flag=True, show_default=True, default=False, help="Suppress logging to stderr.")
@click.option("--log/--no_log", is_flag=True, show_default=True, default=True, help="Log to file.")
@click.option("--progress", is_flag=True, show_default=True, default=False, help="Show a progress bar.")
@click.option("--first_n", default=DEFAULT_FIRST_N, help="Process only this many records. [default: all]")
@click.option(
    "--warnings_as_errors", "-e", is_flag=True, show_default=True, default=False, help="Warnings cause exceptions."
)
@click.version_option(version=cli_builder.version, prog_name=PROGRAM_NAME)
@cli_builder.init_dual_logger()
@cli_builder.init_user_context_obj(extra_args=["progress", "first_n"])
def cli(verbose, quiet, log, **kwargs):
    """bionorm -- normalize and verify genomic data files.

    If COMMAND is present, and --no_log was not invoked,
    a log file named bionorm-COMMAND.log
    will be written in the ./logs/ directory.
    """
    bionorm_check(**kwargs)


cli_builder.set_cli_func(cli)

# Define some cli-related commands
test_log = cli_builder.test_log_func()
show_context = cli_builder.show_context_func()

# import other CLI functions
from .prefix import prefix_fasta  # isort:skip
from .prefix import prefix_gff  # isort:skip
from .extract_fasta import extract_fasta  # isort:skip



# -*- coding: utf-8 -*-
"""Define a cli with dual logging to stderr and file
"""

# standard library imports
import functools
import logging
import sys
from datetime import datetime
from pathlib import Path

# third-party imports
import click

try:
    from importlib.metadata import distribution  # python 3.8 and above
except ModuleNotFoundError:
    from importlib_metadata import distribution


# global constants
DEFAULT_FILE_LOGLEVEL = logging.DEBUG
DEFAULT_STDERR_LOGLEVEL = logging.INFO
STARTTIME = datetime.now()


class CleanInfoFormatter(logging.Formatter):
    """very clean logging formatter for terminal output
    """

    def __init__(self, fmt="%(levelname)s: %(message)s"):
        logging.Formatter.__init__(self, fmt)

    def format(self, record):
        if record.levelno == logging.INFO:
            return record.getMessage()
        return logging.Formatter.format(self, record)


def composed(self, *decs):
    def deco(f):
        for dec in reversed(decs):
            f = dec(f)
        return f

    return deco


def click_multi(func):
    return composed(
        click.option("--progress", is_flag=True, show_default=True, default=False, help="Show a progress bar.")
        # click.option(*self.global_options_list[0]['args'],
        #             **self.global_options_list[0]['kwargs'])
    )(func)


class Logging_CLI_Builder(object):
    """creates a cli function
    """

    def __init__(self, name, logger, check_func=None, global_options_list=None):
        self.name = name
        self.logger = logger
        self.check_func = check_func
        self.global_options_list = global_options_list
        dist = distribution(name).metadata
        self.version = dist["Version"]
        self.author = dist["Author"]
        self.email = dist["Author-email"]
        self.maintainer = dist["Maintainer"]
        self.maintainer_email = dist["Maintainer-email"]
        self.home = dist["Home-page"]
        self.summary = dist["Summary"]
        self.licence = dist["License"]
        self.copyright = "Copyright (C) 2020. National Center for Genome Resources. All rights reserved."
        self.cli_func = None

    def set_cli_func(self, cli_func):
        self.cli_func = cli_func

    def _ctx(self):
        """private global context function"""
        return click.get_current_context

    def get_user_context_dict(self):
        """Returns the user context dictionary.
        """
        return self._ctx()().obj

    def init_user_context_obj(self, extra_args=None):
        """Decorator that puts global options into context dictionary
        """

        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                self._ctx()().obj = {}
                ctx_dict = self._ctx()().obj
                if self._ctx()().params["verbose"]:
                    ctx_dict["logLevel"] = "verbose"
                elif self._ctx()().params["quiet"]:
                    ctx_dict["logLevel"] = "quiet"
                else:
                    ctx_dict["logLevel"] = "default"
                for key in extra_args:
                    ctx_dict[key] = self._ctx()().params[key]
                return f(*args, **kwargs)

            return wrapper

        return decorator

    def init_dual_logger(self, file_log_level=DEFAULT_FILE_LOGLEVEL, stderr_log_level=DEFAULT_STDERR_LOGLEVEL):
        """Decorator to log to stderr and to logfile at different levels
            """

        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                # find the verbose/quiet levels from context
                if self._ctx()().params["verbose"]:
                    _log_level = logging.DEBUG
                elif self._ctx()().params["quiet"]:
                    _log_level = logging.ERROR
                else:
                    _log_level = stderr_log_level
                self.logger.setLevel(file_log_level)
                stderrHandler = logging.StreamHandler(sys.stderr)
                stderrFormatter = CleanInfoFormatter()
                stderrHandler.setFormatter(stderrFormatter)
                stderrHandler.setLevel(_log_level)
                self.logger.addHandler(stderrHandler)
                if self._ctx()().params["log"]:  # start a log file
                    # If a subcommand was used, log to a file in the
                    # logs/ subdirectory of the current working directory
                    #  with the subcommand in the file name.
                    subcommand = self._ctx()().invoked_subcommand
                    if subcommand is not None:
                        logfile_name = self.name + "-" + subcommand + ".log"
                        logfile_path = Path("./logs/" + logfile_name)
                        if not logfile_path.parent.is_dir():  # create logs/ dir
                            try:
                                logfile_path.parent.mkdir(mode=0o755, parents=True)
                            except OSError:
                                self.logger.error('Unable to create logfile directory "%s"', logfile_path.parent)
                                raise OSError
                        else:
                            if logfile_path.exists():
                                try:
                                    logfile_path.unlink()
                                except OSError:
                                    self.logger.error('Unable to remove existing logfile "%s"', logfile_path)
                                    raise OSError
                        logfileHandler = logging.FileHandler(str(logfile_path))
                        logfileFormatter = logging.Formatter("%(levelname)s: %(message)s")
                        logfileHandler.setFormatter(logfileFormatter)
                        logfileHandler.setLevel(file_log_level)
                        self.logger.addHandler(logfileHandler)
                self.logger.debug('Command line: "%s"', " ".join(sys.argv))
                self.logger.debug("%s version %s", self.name, self.version)
                self.logger.debug("Run started at %s", str(STARTTIME)[:-7])
                return f(*args, **kwargs)

            return wrapper

        return decorator

    def log_elapsed_time(self):
        """Decorator to log the elapsed time for command
        """

        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                returnobj = f(*args, **kwargs)
                self.logger.debug("Elapsed time is %s", str(datetime.now() - STARTTIME)[:-7])
                return returnobj

            return wrapper

        return decorator

    def test_log_func(self):
        @self.cli_func.command()
        @self.log_elapsed_time()
        def test_logging():
            """Log at different severity levels.
            """
            self.logger.debug("debug message")
            self.logger.info("info message")
            self.logger.warning("warning message")
            self.logger.error("error message")

    def show_context_func(self):
        @self.cli_func.command()
        def show_context_dict():
            """Print the global context dictionary.
            """
            user_ctx = self.get_user_context_dict()
            self.logger.info("User context dictionary:")
            for key in user_ctx.keys():
                self.logger.info("   %s: %s", key, user_ctx[key])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bionorm']

package_data = \
{'': ['*'], 'bionorm': ['templates/*']}

install_requires = \
['biopython>=1.76,<2.0',
 'click>=7.0,<8.0',
 'click_plugins>=1.1.1,<2.0.0',
 'importlib_metadata>=1.5.0,<2.0.0',
 'ruamel.yaml>=0.16.6,<0.17.0',
 'sequencetools>=0.0.5,<0.0.6']

entry_points = \
{'console_scripts': ['bionorm = bionorm:cli']}

setup_kwargs = {
    'name': 'bionorm',
    'version': '0.2.4',
    'description': 'normalize and verify genomic data',
    'long_description': "bionorm\n=======\n``bionorm`` normalizes and validates genomic data files prior to further processing\nor inclusion in a data store such as that of the\n`Legume Federation <https://www.legumefederation.org/en/data-store/>`_.\n\nPrerequisites\n-------------\nPython 3.6 or greater is required.\nThis package is tested under Linux and MacOS using Python 3.7.\n\nInstallation for Users\n----------------------\nInstall via pip or (better yet) `pipx <https://pipxproject.github.io/pipx/>`_: ::\n\n     pipx install bionorm\n\nFor Developers\n--------------\nIf you plan to develop ``bionorm``, you'll need to install\nthe `poetry <https://python-poetry.org>`_ dependency manager.\nIf you haven't previously installed ``poetry``, execute the command: ::\n\n    curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python\n\nNext, get the master branch from GitHub ::\n\n\tgit clone https://github.com/ncgr/bionorm.git\n\nChange to the ``bionorm/`` directory and install with poetry: ::\n\n\tpoetry install -v\n\nRun ``bionorm`` with ``poetry``: ::\n\n    poetry run bionorm\n\nUsage\n-----\nInstallation puts a single script called ``bionorm`` in your path.  The usage format is::\n\n    bionorm [GLOBALOPTIONS] COMMAND [COMMANDOPTIONS][ARGS]\n\nGlobal Options\n--------------\nThe following options are global in scope and, if used, must be placed before\n``COMMAND``. Not all commands support every global option:\n\n============================= ====================================================\n    --verbose/-v              Log debugging info to stderr.\n    --quiet/-q                Suppress logging to stderr.\n    --no_log                  Suppress logging to file.\n    --progress                Show a progress bar.\n    --first_n                 Process only this many records. [default: all]\n    --warnings_as_errors/-e   Warnings cause exceptions.\n============================= ====================================================\n\nCommands\n--------\nA listing of commands is available via ``bionorm --help``.\nThe currently implemented commands are:\n\n============================= ====================================================\n  prefix_fasta                Prefix FASTA files for data store standard.\n  prefix_gff                  Prefix and sort GFF3 file for data store standard.\n  busco                       Perform BUSCO checks.\n  detector                    Detect/correct incongruencies among files.\n  fasta                       Check for GFF/FASTA consistency.\n  generate_readme             Generates a README file with details of genome.\n  index                       Indexes FASTA file.\n============================= ====================================================\n\nEach command has its ``COMMANDOPTIONS``, which may be listed with: ::\n\n    bionorm COMMAND --help\n\nProject Status\n--------------\n+-------------------+------------+------------+\n| Latest Release    | |pypi|     | |bionorm|  |\n+-------------------+------------+            +\n| GitHub            | |repo|     |            |\n+-------------------+------------+            +\n| License           | |license|  |            |\n+-------------------+------------+            +\n| Travis Build      | |travis|   |            |\n+-------------------+------------+            +\n| Coverage          | |coverage| |            |\n+-------------------+------------+            +\n| Code Grade        | |codacy|   |            |\n+-------------------+------------+            +\n| Dependencies      | |depend|   |            |\n+-------------------+------------+            +\n| Pre-commit        | |precommit||            |\n+-------------------+------------+            +\n| Issues            | |issues|   |            |\n+-------------------+------------+------------+\n\n.. |bionorm| image:: docs/normal.jpg\n     :alt: Make me NORMAL, please!\n\n.. |pypi| image:: https://img.shields.io/pypi/v/bionorm.svg\n    :target: https://pypi.python.org/pypi/bionorm\n    :alt: Python package\n\n.. |repo| image:: https://img.shields.io/github/commits-since/ncgr/bionorm/0.1.0.svg\n    :target: https://github.com/ncgr/bionorm\n    :alt: GitHub repository\n\n.. |license| image:: https://img.shields.io/badge/License-BSD%203--Clause-blue.svg\n    :target: https://github.com/ncgr/bionorm/blob/master/LICENSE.txt\n    :alt: License terms\n\n.. |travis| image:: https://img.shields.io/travis/ncgr/bionorm.svg\n    :target:  https://travis-ci.org/ncgr/bionorm\n    :alt: Travis CI\n\n.. |codacy| image:: https://api.codacy.com/project/badge/Grade/b23fc0c167fc4660bb649320e14dac7f\n    :target: https://www.codacy.com/gh/ncgr/bionorm?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=ncgr/bionorm&amp;utm_campaign=Badge_Grade\n    :alt: Codacy.io grade\n\n.. |coverage| image:: https://codecov.io/gh/ncgr/bionorm/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/ncgr/bionorm\n    :alt: Codecov.io test coverage\n\n.. |precommit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n    :target: https://github.com/pre-commit/pre-commit\n    :alt: pre-commit\n\n.. |issues| image:: https://img.shields.io/github/issues/ncgr/bionorm.svg\n    :target:  https://github.com/ncgr/bionorm/issues\n    :alt: Issues reported\n\n\n.. |depend| image:: https://api.dependabot.com/badges/status?host=github&repo=ncgr/bionorm\n     :target: https://app.dependabot.com/accounts/ncgr/repos/236847525\n     :alt: dependabot dependencies\n",
    'author': 'Connor Cameron',
    'author_email': 'ctc@ncgr.org',
    'maintainer': 'Joel Berendzen',
    'maintainer_email': 'joelb@ncgr.org',
    'url': 'https://github.com/ncgr/ncgr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
